// Benchmark harness for SAXPY/DAXPY
// #define DOUBLE for DAXPY. Otherwise, SAXPY is used

#ifdef DOUBLE
#define AXPY   daxpy
#define FLOAT  double
#else
#define AXPY   saxpy
#define FLOAT  float
#endif

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h> 
#include <unistd.h>
#include <assert.h>
#define NTOTAL      260000000
#define MAX_RECORDS 268435456

#define __USE_GNU
#include <sched.h>

static void print_performance(struct timespec begin, struct timespec end) {
  const long diff =
      (end.tv_sec - begin.tv_sec) * 1000 * 1000 + 
      ((end.tv_nsec - begin.tv_nsec) / 1000);
  printf("%.3g ms\n", (double)diff / 1000.);
}
static struct timespec cur_time(void) {
  struct timespec t;
  assert(clock_gettime(CLOCK_THREAD_CPUTIME_ID,&t) == 0);
  return t;
}

// From
// https://stackoverflow.com/questions/45472147/lost-cycles-on-intel-an-inconsistency-between-rdtsc-and-cpu-clk-unhalted-ref-ts
// https://github.com/travisdowns/turbo-cycles-mystery
// It is claimed that the loop executes one cycle per iteration on recent
// hardware
void busy_loop(uint64_t iters) {
    volatile int sink;
    do {
        sink = 0;
    } while (--iters > 0);
    (void)sink;
}

// To see if the benchmark is running on the same CPU and at the same
// CPU frequency
// The program below estimates the CPU frequency in Mhz
// <skip>using the TSC
// counter (which, on recent hardware), counts time rather than cycles.
// Or, to be precise, it counts reference cycles
// </skip>
// Invoking this function has the side-effect of warming-up CPU
// bringing it from any sleepy or Pn state into the main P0 state
// and perhaps ramping up the frequency if the turbo-boost is enabled.
// (For really running the benchmark, one better disable turbo-boost).
// One has to always keep in mind that the OS scheduler may move the
// process to a different CPU at any time (and back)
// Therefore, the reported information is best effort and may be stale
// Still, it is useful as a ballpark estimate.
static float get_cpu_freq(void) {
  int const calibration_loops = 30 * 1000 * 1000;

  const struct timespec begin = cur_time();
  busy_loop(calibration_loops);
  const struct timespec end = cur_time();

  return 1000. * (double)calibration_loops / (double)(
      (end.tv_sec - begin.tv_sec) * 1000L * 1000L * 1000L + 
      (end.tv_nsec - begin.tv_nsec));
}

// NB
// #include <x86intrin.h>
// int64_t tsc =__rdtsc();
// But one should include __mm_lfence or cpuid before and after rdtsc,
// to serialize


// creating test vectors
static void 
init(const int ntotal, FLOAT ** x, FLOAT ** y) { 

  *x = (FLOAT *)malloc(MAX_RECORDS * sizeof(FLOAT));
  *y = (FLOAT *)malloc(MAX_RECORDS * sizeof(FLOAT));

  srand48(17); // make reproducible
  for (int i = 0; i < ntotal; i++) {
    (*x)[i] = drand48();
    (*y)[i] = drand48();
    }
}

// prototype: external
void AXPY(const int N, const FLOAT da, 
          const FLOAT x[], const int inc_x, FLOAT y[], const int inc_y);


static double average(const int n, const FLOAT x[], const int inc_x) {
  assert (n>1);
  double sum = 0;
  for(int i=0; i<n; i++)
    sum += x[i*inc_x];
  return sum / (double)n;
}


// arg 1: increment (both incr_x and incr_y)
// arg 2: array size 
int main(int argc, const char *argv[]) {

  assert(argc >= 1 && argc <= 3);

  int ntotal = NTOTAL;
  int incr = 1;
  if (argc > 1)
    assert(sscanf(argv[1],"%d",&incr) == 1);
  if (argc > 2)
    assert(sscanf(argv[2],"%d",&ntotal) == 1);
  assert(ntotal < MAX_RECORDS);

  FLOAT *x, *y;
  init(ntotal,&x,&y);

  const int n = ntotal / incr; 

  const int cpu_beg = sched_getcpu();
  const float freq_beg = get_cpu_freq();
  const struct timespec begin = cur_time();
  AXPY(n,2.0,x,incr,y,incr);
  const struct timespec end = cur_time();
  const int cpu_end = sched_getcpu();
  const float freq_end = get_cpu_freq();

  // just for verification
  // the print-out is tuned for turning into space-separated file for gnuplot
  printf("size inc n Avg (begin CPU%d@%8.1f MHz, end CPU%d@%8.1f MHz)\n"
         "%d %3d %-9d %g ", 
         cpu_beg,freq_beg,cpu_end,freq_end,
         ntotal, incr, n, average(n,y,incr));
  print_performance(begin, end);
  return 0;
}
